/**************************************************************************
*   Copyright (C) 2005 by Achal Dhir                                      *
*   achaldhir@gmail.com                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

// tftpserver.cpp

#include <sys/types.h>
#include <limits.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>
#include <memory.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string>
#include <pthread.h>
#include <unistd.h>

#ifndef LOG_MAKEPRI
#define	LOG_MAKEPRI(fac, pri)	(((fac) << 3) | (pri))
#endif

//Constants
#define MAX_BLOCK_SIZE 4096
#define WORD unsigned short
#define BYTE unsigned char
#define DWORD unsigned long

struct ack
{
	WORD opcode;
	union
	{
		WORD block;
		char buffer[510];
	};
};

struct packet
{
	WORD opcode;
	WORD block;
	char buffer[MAX_BLOCK_SIZE];
};

struct tftperror
{
	WORD opcode;
	WORD errorcode;
	char errormessage[508];
};

struct request
{
	timeval tv;
	fd_set readfds;
	pthread_t threadId;
	BYTE attempt;
	char path[256];
	FILE *file;
	char *filename;
	char *mode;
	DWORD tsize;
	DWORD blksize;
	WORD interval;
	WORD block;
	int bytesRecd;
	int bytesRead[2];
	int bytesSent;
	sockaddr_in client;
	socklen_t clientsize;
	packet pkt[2];
	tftperror er;
	ack acout;
	packet datain;
	int knock;
};

//Functions
void runProg();
void *ProcessRequest(void *lpParam);
char *cleanstr(char* buff, bool next);
void initialize();
char *myUpper(char *string);
char *myLower(char *string);
bool getSection(char *sectionName, char *buffer, int sizeofbuffer, char *fileName);


//Global Variables
bool verbatim = false;
char homeDir[512]="/home/";
char iniFile[128]="/etc/tftpserver.ini";
WORD blksize = MAX_BLOCK_SIZE;
WORD interval = 5;
char overwrite = false;
DWORD serverip = INADDR_ANY;
int minport = 0;
int maxport = 0;

int main(int argc, char **argv)
{
	verbatim = (argc > 1 && !strcasecmp(argv[1], "-v"));
	if (verbatim)
	{
		if (getuid())
		{
			printf("Error: Only root should run this program\n");
			exit(1);
		}

		initialize();
		verbatim = true;

		// Create a socket.
		int m_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

		if (m_socket == -1)
		{
			printf( "socket creation failed.\n" );
			return 1;
		}

		// Bind the socket.
		sockaddr_in service;
		service.sin_family = AF_INET;
		service.sin_addr.s_addr = serverip;
		service.sin_port = htons(69);

		if ( bind(m_socket, (sockaddr*) &service, sizeof(service) ) == -1)
		{
			if (serverip)
				printf( "Invalid listen address %s or Port 69 already in use.\n",inet_ntoa(service.sin_addr));
			else
				printf( "Port 69 already in use.\n");

			close(m_socket);
			return 1;
		}

		printf("Home Directory: %s\n", homeDir);
		printf("Listening On: %s\n", inet_ntoa(service.sin_addr));

		if (minport)
			printf("port range: %u-%u\n", minport, maxport);

		printf("max blksize: %u\n", blksize);
		printf("defult blksize: %u\n", 512);
		printf("default interval: %u\n", interval);
		printf("Overwrite Existing Files: %s\n", overwrite ? "Yes" : "No");
		printf( "\nReady...\n" );

		timeval tv;
		fd_set readfds;

		do
		{
			tv.tv_sec = 20;
			tv.tv_usec = 0;
			FD_ZERO(&readfds);
			FD_SET(m_socket, &readfds);
			select(USHRT_MAX, &readfds, 0, 0, &tv);
			if (FD_ISSET(m_socket, &readfds))
			{
				//printf("knock\n");
				request *req = (request*)malloc(sizeof(request));
				memset(req, 0, sizeof(request));
				req->clientsize = sizeof(req->client);
				errno = 0;
				req->bytesRecd = recvfrom(m_socket, (char*) & req->datain, sizeof(packet), 0, (sockaddr*) & req->client, &req->clientsize);
				req->knock = m_socket;
				if (!errno)
				{
					if (req->bytesRecd > 0 && (ntohs(req->datain.opcode) == 1 || ntohs(req->datain.opcode) == 2))
					{
						pthread_create(&req->threadId, 0, ProcessRequest, (void*)req);
						if (!req->threadId)
						{
							req->er.opcode = htons(5);
							req->er.errorcode = htons(0);
							strcpy(req->er.errormessage, "Thread Creation Error");
							req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
							printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);
							free(req);
						}
					}
					else
					{
						req->er.opcode = htons(5);
						req->er.errorcode = htons(0);
						sprintf(req->er.errormessage, "Invalid Option Code %i", ntohs(req->datain.opcode));
						req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
						printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);
						free(req);
					}
				}
				else
					free(req);
			}
		}
		while (true);
		close(m_socket);
	}
	else
	{
		if (getuid())
		{
			syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), "Only root should run this program");
			exit(1);
		}

		/* Our process ID and Session ID */
		pid_t pid, sid;

		/* Fork off the parent process */
		pid = fork();
		if (pid < 0)
		{
			exit(EXIT_FAILURE);
		}
		/* If we got a good PID, then
		we can exit the parent process. */
		if (pid > 0)
		{
			exit(EXIT_SUCCESS);
		}

		/* Change the file mode mask */
		umask(0);

		/* Open any logs here */

		/* Create a new SID for the child process */
		sid = setsid();
		if (sid < 0)
		{
			/* Log the failure */
			exit(EXIT_FAILURE);
		}

		/* Close out the standard file descriptors */
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);

		/* Daemon-specific initialization goes here */
		//Initialize
		initialize();
		verbatim = false;

		// Create a socket.
		int m_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

		if (m_socket == -1)
		{
			return 1;
		}

		// Bind the socket.
		sockaddr_in service;
		service.sin_family = AF_INET;
		service.sin_addr.s_addr = serverip;
		service.sin_port = htons(69);

		if ( bind(m_socket, (sockaddr*) &service, sizeof(service) ) == -1)
		{
			if (serverip)
				syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), "Invalid listen address or Port 69 already in use");
			else
				syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), "Port 69 already in use");

			close(m_socket);
			return 1;
		}

		timeval tv;
		fd_set readfds;

		/* The Big Loop */
		do
		{
			tv.tv_sec = 20;
			tv.tv_usec = 0;
			FD_ZERO(&readfds);
			FD_SET(m_socket, &readfds);
			select(USHRT_MAX, &readfds, 0, 0, &tv);
			if (FD_ISSET(m_socket, &readfds))
			{
				//printf("knock\n");
				request *req = (request*)malloc(sizeof(request));
				memset(req, 0, sizeof(request));
				req->clientsize = sizeof(req->client);
				errno = 0;
				req->bytesRecd = recvfrom(m_socket, (char*) & req->datain, sizeof(packet), 0, (sockaddr*) & req->client, &req->clientsize);
				req->knock = m_socket;
				if (!errno)
				{
					if (req->bytesRecd > 0 && (ntohs(req->datain.opcode) == 1 || ntohs(req->datain.opcode) == 2))
					{
						pthread_create(&req->threadId, 0, ProcessRequest, (void*)req);
						if (!req->threadId)
						{
							req->er.opcode = htons(5);
							req->er.errorcode = htons(0);
							strcpy(req->er.errormessage, "Thread Creation Error");
							req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
							free(req);
						}
					}
					else
					{
						req->er.opcode = htons(5);
						req->er.errorcode = htons(0);
						sprintf(req->er.errormessage, "Invalid Option Code %i", ntohs(req->datain.opcode));
						req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
						free(req);
					}
				}
				else
					free(req);
			}
		}
		while (true);
		close(m_socket);
	}
}

void *ProcessRequest(void *lpParam)
{
	//printf("here\n");
	request *req = (request*)lpParam;
	req->blksize = 512;
	req->interval = interval;
	bool fetchAck = false;

	int m_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (m_socket == -1)
	{
		req->er.opcode = htons(5);
		req->er.errorcode = htons(0);
		strcpy(req->er.errormessage, "Thread Socket Creation Error");
		req->bytesSent = sendto(req->knock, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) &req->client, req->clientsize);

		if (verbatim)
			printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

		free(req);
		pthread_exit(NULL);
	}

	//int bOptVal = true;
	//int bOptLen = sizeof(bOptVal);
	//setsockopt(m_socket, SOL_SOCKET, SO_REUSEADDR, (char*)&bOptVal, bOptLen);

	//int bOptVal = true;
	//int bOptLen = sizeof(bOptVal);
	//setsockopt(m_socket, SOL_SOCKET, SO_DONTLINGER, (char*)&bOptVal, bOptLen);

	// Bind the socket.
	WORD comport;
	sockaddr_in service;

	for (comport = minport;;comport++)
	{
		service.sin_family = AF_INET;
		service.sin_addr.s_addr = serverip;
		service.sin_port = htons(comport);

		if (comport > maxport)
		{
			req->er.opcode = htons(5);
			req->er.errorcode = htons(0);
			strcpy(req->er.errormessage, "No port is free");
			req->bytesSent = sendto(req->knock, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);

			if (verbatim)
				printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

			close(m_socket);
			free(req);
			pthread_exit(NULL);
		}
		else if (bind(m_socket, (sockaddr*) &service, sizeof(service) ) == -1)
			continue;
		else
			break;
	}

	if (connect(m_socket, (sockaddr*)&req->client, req->clientsize) == -1)
	{
		req->er.opcode = htons(5);
		req->er.errorcode = htons(0);
		strcpy(req->er.errormessage, "Connect Failed");
		req->bytesSent = sendto(req->knock, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);

		if (verbatim)
			printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

		close(m_socket);
		free(req);
		pthread_exit(NULL);
	}

	char *temp = (char*) & req->datain;
	temp += 2;
	req->filename = temp;
	temp += strlen(temp) + 1;
	req->mode = temp;
	temp += strlen(temp) + 1;
	strcpy(req->path, homeDir);
	strcat(req->path, req->filename);

	for (int i = 0; i < strlen(req->path); i++)
		if (req->path[i] == '\\')
			req->path[i] = '/';

	if (strstr(req->path, "../"))
	{
		req->er.opcode = htons(5);
		req->er.errorcode = htons(2);
		strcpy(req->er.errormessage, "Access violation");

		if (verbatim)
			printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

		req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
		close(m_socket);
		free(req);
		pthread_exit(NULL);
	}

	if (ntohs(req->datain.opcode) == 1)
	{
		errno = 0;

		if (!strcasecmp(req->mode, "netascii") || !strcasecmp(req->mode, "ascii"))
			req->file = fopen(req->path, "rt");
		else
			req->file = fopen(req->path, "rb");

		if (errno || !req->file)
		{
			req->er.opcode = htons(5);
			req->er.errorcode = htons(1);
			strcpy(req->er.errormessage, "File Not Found");

			if (verbatim)
				printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

			req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
			close(m_socket);
			free(req);
			pthread_exit(NULL);
		}
	}
	else
	{
		if (!overwrite)
		{
			req->file = fopen(req->path, "rb");
			if (req->file)
			{
				req->er.opcode = htons(5);
				req->er.errorcode = htons(6);
				strcpy(req->er.errormessage, "File already exists");

				if (verbatim)
					printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

				req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
				close(m_socket);
				free(req);
			pthread_exit(NULL);
			}
		}

		errno = 0;

		if (!strcasecmp(req->mode, "netascii") || !strcasecmp(req->mode, "ascii"))
			req->file = fopen(req->path, "wt");
		else
			req->file = fopen(req->path, "wb");

		if (errno || !req->file)
		{
			req->er.opcode = htons(5);
			req->er.errorcode = htons(2);
			strcpy(req->er.errormessage, "Invalid Path");

			if (verbatim)
				printf("Client %s:%u %s, %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

			req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
			close(m_socket);
			free(req);
			pthread_exit(NULL);
		}
	}

	if (*temp)
	{
		fetchAck = true;
		char *pointer = req->acout.buffer;
		req->acout.opcode = htons(6);
		DWORD val;
		while (*temp)
		{
			//printf("%s", temp);
			if (!strcasecmp(temp, "blksize"))
			{
				strcpy(pointer, temp);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;
				val = atol(temp);

				if (val > 512 && val <= blksize)
					req->blksize = val;

				sprintf(pointer, "%u", req->blksize);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;

			}
			else if (!strcasecmp(temp, "tsize"))
			{
				strcpy(pointer, temp);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;
				fseek(req->file, 0, SEEK_END);
				req->tsize = ftell(req->file);
				sprintf(pointer, "%u", req->tsize);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;
			}
			else if (!strcasecmp(temp, "interval"))
			{
				strcpy(pointer, temp);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;
				val = atoi(temp);

				if (val > 0 && val <= 100)
				{
					req->interval = val;
				}

				sprintf(pointer, "%u", req->interval);
				pointer += strlen(pointer) + 1;
				temp += strlen(temp) + 1;
			}
			else
				temp += strlen(temp) + 1;

			//printf("=%u\n",val);
		}

		req->bytesSent = sendto(m_socket, (const char*) & req->acout, (DWORD)pointer - (DWORD) & req->acout, 0, (sockaddr*) & req->client, req->clientsize);
	}
	else if (htons(req->datain.opcode) == 2)
	{
		req->acout.opcode = htons(4);
		req->acout.block = htons(0);
		req->bytesSent = sendto(m_socket, (const char*) & req->acout, 4, 0, (sockaddr*) & req->client, req->clientsize);
	}

	if (errno)
	{
		req->er.opcode = htons(5);
		req->er.errorcode = htons(0);
		sprintf(req->er.errormessage, strerror(errno));
		req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);

		if (verbatim)
			printf("Client %s:%u %s, error: %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);

		req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);

		if (req->file)
			fclose(req->file);

		close(m_socket);
		free(req);
		pthread_exit(NULL);
	}
	else if (ntohs(req->datain.opcode) == 1)
	{
		if (ftell(req->file))
			fseek(req->file, 0, SEEK_SET);

		req->pkt[0].opcode = htons(3);
		req->pkt[0].block = htons(1);
		req->bytesRead[0] = fread(req->pkt[0].buffer, 1, req->blksize, req->file);

		if (errno)
		{
			req->er.opcode = htons(5);
			req->er.errorcode = htons(0);
			sprintf(req->er.errormessage, "%s", strerror(errno));
			if (verbatim)
				printf("Client %s:%u %s, error: %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);
			req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
			if (req->file)
				fclose(req->file);

			close(m_socket);
			free(req);
			pthread_exit(NULL);
		}

		if (req->bytesRead[0] == req->blksize)
		{
			req->pkt[1].opcode = htons(3);
			req->pkt[1].block = htons(2);
			req->bytesRead[1] = fread(req->pkt[1].buffer, 1, req->blksize, req->file);
			if (req->bytesRead[1] < req->blksize)
			{
				fclose(req->file);
				req->file = 0;
			}
		}
		else
		{
			fclose(req->file);
			req->file = 0;
		}

		while (true)
		{
			errno = 0;
			if (req->attempt > 3)
			{
				req->er.opcode = htons(5);
				req->er.errorcode = htons(0);
				strcpy(req->er.errormessage, "Timeout");
				break;
			}

			if (fetchAck)
			{
				req->tv.tv_sec = req->interval;
				req->tv.tv_usec = 0;
				FD_ZERO(&req->readfds);
				FD_SET(m_socket, &req->readfds);
				select(USHRT_MAX, &req->readfds, NULL, NULL, &req->tv);
				if (FD_ISSET(m_socket, &req->readfds))
				{
					req->bytesRecd = recvfrom(m_socket, (char*) & req->datain, sizeof(packet), 0, (sockaddr*) & req->client, &req->clientsize);
					if (errno)
						break;
				}
				else
					req->bytesRecd = 0;
			}
			else
				req->datain.block = htons(0);

			fetchAck = true;

			//printf("%i\n",ntohs(req->datain.opcode));

			if (req->bytesRecd >= 4)
			{
				if (ntohs(req->datain.opcode) == 1 || ntohs(req->datain.opcode) == 4)
				{
					if (ntohs(req->datain.block) == req->block)
					{
						req->block++;
						req->attempt = 0;
					}
					else
						continue;

					if (req->block == USHRT_MAX)
					{
						req->er.opcode = htons(5);
						req->er.errorcode = htons(3);
						strcpy(req->er.errormessage, "File Too Large");
						break;
					}
				}
				else if (ntohs(req->datain.opcode) == 5)
				{
					memcpy(&req->er, &req->datain, req->bytesRecd);
					break;
				}
				else
				{
					req->er.opcode = htons(5);
					req->er.errorcode = htons(0);
					strcpy(req->er.errormessage, "Unexpected Option Code");
					break;
				}
			}
			else
				req->attempt++;

			if (ntohs(req->pkt[0].block) == req->block)
			{
				req->bytesSent = sendto(m_socket, (const char*) & req->pkt[0], req->bytesRead[0] + 4, 0, (sockaddr*) & req->client, req->clientsize);
				if (errno)
					break;
				if (req->file)
				{
					if (ntohs(req->pkt[1].block) < ntohs(req->pkt[0].block))
					{
						req->pkt[1].block = htons(req->block + 1);
						req->bytesRead[1] = fread(req->pkt[1].buffer, 1, req->blksize, req->file);
						if (req->bytesRead[1] < req->blksize)
						{
							fclose(req->file);
							req->file = 0;
						}
					}
				}
			}
			else if (ntohs(req->pkt[1].block) == req->block)
			{
				req->bytesSent = sendto(m_socket, (const char*) & req->pkt[1], req->bytesRead[1] + 4, 0, (sockaddr*) & req->client, req->clientsize);
				if (errno)
					break;
				if (req->file)
				{
					if (ntohs(req->pkt[0].block) < ntohs(req->pkt[1].block))
					{
						req->pkt[0].block = htons(req->block + 1);
						req->bytesRead[0] = fread(req->pkt[0].buffer, 1, req->blksize, req->file);
						if (req->bytesRead[0] < req->blksize)
						{
							fclose(req->file);
							req->file = 0;
						}
					}
				}
			}
			else if (ntohs(req->pkt[0].block) < req->block && ntohs(req->pkt[1].block) < req->block)
			{
				break;
			}
		}

		if (verbatim && !errno && !req->er.errormessage[0])
			printf("Client %s:%u %s, %u Blocks Served\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->block-1);
	}
	else if (ntohs(req->datain.opcode) == 2)
	{
		//BYTE reSends = 0;
		while (true)
		{
			if (req->attempt > 3)
			{
				req->er.opcode = htons(5);
				req->er.errorcode = htons(0);
				strcpy(req->er.errormessage, "Timeout");
				break;
			}

			req->tv.tv_sec = req->interval;
			req->tv.tv_usec = 0;
			FD_ZERO(&req->readfds);
			FD_SET(m_socket, &req->readfds);
			select(USHRT_MAX, &req->readfds, NULL, NULL, &req->tv);

			errno = 0;
			if (FD_ISSET(m_socket, &req->readfds))
			{
				req->bytesRecd = recvfrom(m_socket, (char*) & req->pkt[0], sizeof(packet), 0, (sockaddr*) & req->client, &req->clientsize);
				if (errno)
					break;
			}
			else
				req->bytesRecd = 0;

			if (req->bytesRecd >= 4)
			{
				if (ntohs(req->pkt[0].opcode) == 3)
				{
					if (ntohs(req->pkt[0].block) == req->block + 1)
					{
						req->acout.opcode = htons(4);
						req->acout.block = req->pkt[0].block;
						req->block++;

						if (req->block == USHRT_MAX)
						{
							req->er.opcode = htons(5);
							req->er.errorcode = htons(3);
							strcpy(req->er.errormessage, "File Too Large");
							break;
						}

						sendto(m_socket, (const char*)&req->acout, 4, 0, (sockaddr*)&req->client, req->clientsize);

						if (errno)
							break;

						if (req->file && req->bytesRecd > 4)
						{
							fwrite(req->pkt[0].buffer, req->bytesRecd - 4, 1, req->file);
							if (errno)
							{
								errno = 0;
								req->er.opcode = htons(5);
								req->er.errorcode = htons(3);
								strcpy(req->er.errormessage, "Disk full or allocation exceeded");
								break;
							}
							else
								req->attempt = 0;
						}

						if (req->bytesRecd - 4 < req->blksize)
						{
							fclose(req->file);
							req->file = 0;
							break;
						}
					}
					else
						continue;
				}
				else if (ntohs(req->pkt[0].opcode) == 5)
				{
					memcpy(&req->er, &req->pkt[0], req->bytesRecd);
					break;
				}
				else
				{
					req->er.opcode = htons(5);
					req->er.errorcode = htons(0);
					strcpy(req->er.errormessage, "Unexpected Option Code");
					break;
				}
			}
			else if (req->file)
			{
				sendto(m_socket, (const char*)&req->acout, req->bytesSent, 0, (sockaddr*)&req->client, req->clientsize);
				req->attempt++;
				if (errno)
					break;
			}
			else
				break;
		}

		if (verbatim && !errno && !req->er.errormessage[0])
			printf("Client %s:%u %s, %u Blocks Received\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->block);
	}

	if (errno || req->er.errormessage[0])
	{
		if (errno)
		{
			req->er.opcode = htons(5);
			req->er.errorcode = htons(0);
			sprintf(req->er.errormessage, "%s", strerror(errno));
		}
		if (verbatim)
			printf("Client %s:%u %s, error: %s\n", inet_ntoa(req->client.sin_addr), ntohs(req->client.sin_port), req->path, req->er.errormessage);
		req->bytesSent = sendto(m_socket, (const char*) & req->er, strlen(req->er.errormessage) + 5, 0, (sockaddr*) & req->client, req->clientsize);
	}

	if (req->file)
		fclose(req->file);

	close(m_socket);
	free(req);
	pthread_exit(NULL);
}

void initialize()
{
	char iniStr[4096];
	if (getSection("TFTP-OPTIONS", iniStr, sizeof(iniStr), iniFile))
	{
		char *iniStrPtr = cleanstr(iniStr, false);
		for (;strlen(iniStrPtr);iniStrPtr = cleanstr(iniStrPtr, true))
		{
			char name[256];
			strncpy(name, iniStrPtr, 255);
			char *value = strchr(name, '=');
			if (value != NULL)
			{
				*value = 0;
				value++;
				myLower(name);
				if (!strcmp(name, "home"))
				{
					strcpy(homeDir, value);
					if (homeDir[strlen(homeDir) - 1] != '/')
						strcat(homeDir, "/");
				}
				else if (!strcmp(name, "listen-on"))
				{
					serverip = inet_addr(value);
					if (serverip == INADDR_NONE)
						serverip = INADDR_ANY;
				}
				else if (!strcmp(name, "blksize"))
				{
					blksize = atol(value);
					if (blksize < 512)
						blksize = 512;
					else if (blksize > MAX_BLOCK_SIZE)
						blksize = MAX_BLOCK_SIZE;
				}
				else if (!strcmp(name, "interval"))
				{
					interval = atol(value);
					if (interval < 1)
						interval = 5;
					else if (interval > 60)
						interval = 60;
				}
				else if (!strcmp(name, "overwrite"))
				{
					if (!strcasecmp(value, "y"))
						overwrite = true;
					else
						overwrite = false;
				}
				else if (!strcmp(name, "port-range"))
				{
					char *ptr = strchr(value,'-');
					if (ptr)
					{
						*ptr = 0;
						minport = atol(value);
						maxport = atol(++ptr);

						if (minport < 1024 || minport >= USHRT_MAX || maxport < 1024 || maxport >= USHRT_MAX || minport > maxport)
						{
							minport = 0;
							maxport = 0;

							if (verbatim)
								printf("Invalid port range %s-%s\n", value, ptr);
						}
					}
					else if (verbatim)
						printf("Invalid port range %s\n", value);
				}
			}
		}
	}
}

char *cleanstr(char* buff, bool next)
{
	if (strlen(buff) && next)
		buff += strlen(buff) + 1;
	while (strlen(buff))
		if (*buff >= 48 && *buff <= 57 || *buff >= 65 && *buff <= 90 || *buff >= 97 && *buff <= 122)
			break;
		else
			buff += strlen(buff) + 1;
	return buff;
}

bool getSection(char *sectionName, char *buffer, int sizeofbuffer, char *fileName)
{
	char section[128];
	sprintf(section, "[%s]", sectionName);
	myUpper(section);
	FILE *f = fopen(fileName, "r");
	char buff[512];
	char *ptr = buffer;
	*ptr = 0;
	bool found = false;
	while (f && fgets(buff, 255, f))
	{
		myUpper(buff);
		if (strstr(buff, section) == buff)
		{
			found = true;
			while (fgets(buff, 255, f))
			{
				if (strstr(buff, "[") == buff)
					break;
				sprintf(ptr, "%s", buff);
				ptr += strlen(buff);
				while (*ptr <= 32)
					ptr--;
				ptr++;
				*ptr = 0;
				ptr++;
			}
			*ptr = 0;
		}
	}
	if (f)
		fclose(f);
	return found;
}

char *myUpper(char *string)
{
	WORD len = strlen(string);
	for (int i = 0; i < len; i++)
		if (string[i] >= 'a' && string[i] <= 'z')
			string[i] -= ('a' - 'A');
	return string;
}

char *myLower(char *string)
{
	WORD len = strlen(string);
	for (int i = 0; i < len; i++)
		if (string[i] >= 'A' && string[i] <= 'Z')
			string[i] += ('a' - 'A');
	return string;
}
